package foo

const GopPackage = true

const Gopo__T__Gop_Add = ".AddInt,AddString"

type T int

// AddInt doc
func (p T) AddInt(b int) *T {
	return nil
}

// AddString doc
func AddString(this *T, b string) *T {
	return nil
}
