/*
 * Copyright (c) 2021 The XGo Authors (xgo.dev). All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

type T struct {
	*T
	A int `json:"a"`
}

func bar(v chan bool) (int, <-chan error) {
	v <- true
	<-v
	return 0, (<-chan error)(nil)
}

func foo(f func([]byte, *string, ...T) chan<- int) (v int, err error) {
	return
}
