// Code generated by xgo (XGo); DO NOT EDIT.

package main

import "fmt"

const _ = true
//line demo/mixgo/b.xgo:1:1
func sayMix() {
//line demo/mixgo/b.xgo:2:1
	fmt.Println("Mix Go and XGo")
}
//line demo/mixgo/b.xgo:5
func main() {
//line demo/mixgo/b.xgo:5:1
	p("world")
}
