package foo

const GopPackage = true

const Gopo_Add = "AddInt,,AddString"

// Add doc
func Add__1(a, b float64) float64 {
	return 0
}

// AddInt doc
func AddInt(a, b int) int {
	return 0
}

// AddString doc
func AddString(a, b string) string {
	return ""
}
